# Cap9 Substrate Contracts

## Overview

The cap9 modifications to the contracts module introduce 3 new functions that
can be imported into contracts.

- `cap9_clist`
- `cap9_clist_downgrade`
- `cap9_call_with_caps`

Each of these new functions act on the two pieces of additional metadata that
have been added to the runtime.

- During a contracts call (be it directly from an external source or from
  another contract) it possesses a set of _contextual_ capabilities. These
  define what can or cannot happen during the execution of this contract.
- Each contract has an _intrinsic_ list of capabilities. These become the
  _contextual_ capabilites whenever a contract is called directly from an
  external source, or via a regular `ext_call`.

### `cap9_clist`

This function has the signature `(func $cap9_clist)`. It takes no arguments and
returns nothing, but fills the scratch buffer with the current contextual
capabilities.

### `cap9_clist_downgrade`

This function has the signature `(func $cap9_clist_downgrade (param i32 i32))`.
Its two arguments are the pointer and length of the memory that contains the
capabilities that will be downgraded to. To _downgrade_ capabilities is to
relinquish some of the intrinsic capabilities that a contract has. When a
contract is instantiated normally it has the full set of capabilities. By using
this function it can voluntarily reduce the capabilities it has. This function
does not affect the current _contextual_ capabilities.

This function is well suited to using during the deploy phase.

### `cap9_call_with_caps`

This function has the signature `(func $cap9_call_with_caps (param i32 i32 i64 i32 i32 i32 i32 i32 i32) (result i32))`. This is like an `ext_call` but with two important differences:

 - The "state" that is used by the called contract is the state of the current
   contract (the one executing this function). That is, if the contract that is
   called with this function writes to storage, it writes to the storage of the
   function that executed this function (much like a delegate call in Ethereum).
 - This function takes a set of capabilities as an argument. This set of
   capabilities defines what the called contract can do. If the set of
   capabilities is zero, it can do nothing but return a value. This set of
   capabilities must also be a subset of the current _contextual_ capabilities,
   i.e. it can do no more than the parent can already do.

The arguments and return value of this function are the same as `ext_call`
except that it has two additional arguments (the last two) which are the pointer
and length (respectively) of the set of capabilities which will be passed.

## Quickstart

First make sure that you've built the substrate node in this directory and have
it running in dev mode. You made need to purge any previous chain data. It's
also wise to split the `stdout` and `stderr`, as the output of our tests will be
on `stdout` and we want to split that from the general node logging. All of this
can be done via the `run` scripts in this directory.

Linux:

```bash
./run.sh
```

Windows:

```cmd
run
```

Both of these files do something along the lines of:

```bash
cargo build --release
cargo run --release -- purge-chain --dev -y
cargo run --release -- --dev > ../substrate.txt 2> ../substrate.err
```

Once this is running, the tests can be run in the `test-harness` directory by
simply executing `cargo run`. This will result in the test output being logged
in substrate.txt.

## Test Overview

The tests, as outlined above, are a little clunky. This is because the current
Substrate version this is based on, and the API client library don't appear to
have the right tools for us to do automated testing. This will likely be fixed
by us changing Substrate versions, but for now the tests must be verified by the
operator of the tests.

The test consists of two contracts: the caller and the callee. The caller sets
up the case, by first storing the value `0x02` at a hardcoded address in
storage. It then calls the callee using `cap9_call_with_caps`, passing in
whichever caps it deems appropriate. During this call, the callee will attempt
to increment the value at this hardcoded storage address by 1 (`0x02` → `0x03`).

If the callee has been passed the appropriate capability, the value should be incremented and logged by the callee and then the caller (as `0x03`). If the callee does not have the appropriate capability it will throw an error (not logging any new value) and return to the caller. The caller will then log the value in storage which should remain as `0x02`.

### The Blocked Case (No Write Cap)

The first case we will look at is the blocking case, where the caller passes the callee a write capability in the arguments to `cap9_call_with_caps`.

Once you have a Substrate node running (using the modified Substrate in this repo) and have executed `cargo run` in the `test-harness` directory, the results of the tests will be outputted to stdout. If you used the runscripts they will be outputted to `substrate.txt`. If you run the tests unmodified you should see something like the output below (the line numbers are added for reference).

```
01: [CALLEE] Caps: 0x0101
02: [CALLEE] Caps: 0x0101
03: DispatchError
04: 16
05: 0
06: during execution
07: [CALLER] Called with: 0x8F424D3D90D24BD260C260DE3D7ED3608E860299417A2CA0C859A56B3349A290
08: [CALLER] Balance: 0x20A1F6EB59AF20000000000000000000
09: [CALLER] Caps: 0x0101
10: [CALLER] Current Storage Value: 0x02
11: [CALLER] Calling From: 0xD407FD44606C528D8BC8F1E31EBC4064A8DFBED3AF804873945FA8F0085FC19F
12: [CALLER] Calling To: 0x8F424D3D90D24BD260C260DE3D7ED3608E860299417A2CA0C859A56B3349A290
13: [CALLEE] Caps: 0x0001
14: [CALLEE] Initial Storage Value: 0x02
15: [CALLER] Return Value: 0x0001
16: [CALLER] Callee threw an error
17: [CALLER] Current Storage Value: 0x02
18: [CALLER] Called with: 0x8F424D3D90D24BD260C260DE3D7ED3608E860299417A2CA0C859A56B3349A290
19: [CALLER] Balance: 0x20A1F6EB59AF20000000000000000000
20: [CALLER] Caps: 0x0101
21: [CALLER] Current Storage Value: 0x02
22: [CALLER] Calling From: 0xD407FD44606C528D8BC8F1E31EBC4064A8DFBED3AF804873945FA8F0085FC19F
23: [CALLER] Calling To: 0x8F424D3D90D24BD260C260DE3D7ED3608E860299417A2CA0C859A56B3349A290
24: [CALLEE] Caps: 0x0001
25: [CALLEE] Initial Storage Value: 0x02
26: [CALLER] Return Value: 0x0001
27: [CALLER] Callee threw an error
28: [CALLER] Current Storage Value: 0x02
29: Encountered an error: connection reset
30: Enable a logger to see more information.
```

Lines 1-6 can be ignored, they are the result of an execution of the callee
contract without the right setup. Lines 18-28 are simply a repetition of 7-17
and can be ignored. Lines 29 & 30 are simply the breaking of the connection at
the end of the test. The lines of interest are 7-17, so they are reproduced here
so we can go through them in more detail.

```
07: [CALLER] Called with: 0x8F424D3D90D24BD260C260DE3D7ED3608E860299417A2CA0C859A56B3349A290
08: [CALLER] Balance: 0x20A1F6EB59AF20000000000000000000
09: [CALLER] Caps: 0x0101
10: [CALLER] Current Storage Value: 0x02
11: [CALLER] Calling From: 0xD407FD44606C528D8BC8F1E31EBC4064A8DFBED3AF804873945FA8F0085FC19F
12: [CALLER] Calling To: 0x8F424D3D90D24BD260C260DE3D7ED3608E860299417A2CA0C859A56B3349A290
13: [CALLEE] Caps: 0x0001
14: [CALLEE] Initial Storage Value: 0x02
15: [CALLER] Return Value: 0x0001
16: [CALLER] Callee threw an error
17: [CALLER] Current Storage Value: 0x02
```

Line 7 is input data of the caller. In this case it is the address of the
callee, which the caller needs to be able to call it. Line 8 is the balance,
which is just for debugging and servers no test purpose here. Line 9 is the
encoded caps. We can see it is `0x0101` (i.e. to bytes of value 1). This
represents to `true` values, for the write and log capabilities respectively. We
can see that the caller has a value of true for both, which means it can write
and log.

Line 10 is the first time we log the storage value for our tests. As you can see
it starts at 0x02. Lines 11 and 12 are simply the addresses of the caller and
callee.

Line 13 is the first log line in the callee (it has just been called). This line
is of particular interest as it shows the current caps of the callee (in this
context) and we can see that unlike the caller, its value for the write
capability is `0x00` or `false`, which means it does not have the capability to
write to storage. If everything is working correctly this means that any write
calls should result in an error, and the execution of the callee being reverted.

Line 14 shows the initial value of the storage location at the start of the call
to callee. As we can see it is also `0x02`. This successfully demonstrates that
the callee is reading the storage values of the caller (which is what we want),
much like a DELEGATECALL in Ethereum.

On line 15 we are back into the logs of the caller. This suggests that the
callee has terminated early, and this is confirmed on the next line where we see
that the exit code of the callee's execution is `0x0001` indicating that the
execution failed and was reverted. Indeed line 16 confirms this for us with
text. This matches with our expectations, as we saw earlier that it did not have
a write capability and the callee is programmed to execute a write call.

Line 18 confirms that the original value of storage remains `0x02`.

### The Unblocked Case (With Write Cap)

The test setup is the same as the blocked case, so only the relevant parts will
be reproduced. In order to run this case it is necessary to change the hardcoded
capability values in the caller. The code for the caller is in `test-harness/src/caller_contract.wat`, and on the last line of code there is capabilities which we will pass to the callee: `(data (i32.const 13000) "\01\01")`. The text `\01\01` is the encoded capabilities, so if we change is to `\00\01` we have removed the write capability, and it will no longer be passed to the callee.

```
07: [CALLER] Called with: 0x8F424D3D90D24BD260C260DE3D7ED3608E860299417A2CA0C859A56B3349A290
08: [CALLER] Balance: 0x20A1F6EB59AF20000000000000000000
09: [CALLER] Caps: 0x0101
10: [CALLER] Current Storage Value: 0x02
11: [CALLER] Calling From: 0xE0E82B61F484990966B9A7A5837E499DDC224117EE80FDF0C1C888B256F81DDC
12: [CALLER] Calling To: 0x8F424D3D90D24BD260C260DE3D7ED3608E860299417A2CA0C859A56B3349A290
13: [CALLEE] Caps: 0x0101
14: [CALLEE] Initial Storage Value: 0x02
15: [CALLEE] Current Storage Value: 0x03
16: [CALLER] Return Value: 0x0000
17: [CALLER] Callee exited successfully
18: [CALLER] Current Storage Value: 0x03
```

Let's start by looking at the first different line, line 13. In this line we can
see that this time the callee has `0x0101` caps, so its write value is set to
`true` and it should be able to write. Indeed if we jump to line 15 (which was
not present last time) we can see that the value has been incremented to `0x03`.
Lines 16 and 17 also indicate that the callee completed and returned without
error.

Most importantly on line 18 we can see that when the caller logs the value from
storage again (after having called the callee) the value has been incremented by
0x03, as expected. This indicates that the tests were successful.
